--Lighting Engine
--KJD Dec 2012

local _lg = love.graphics
local cos = math.cos
local sin = math.sin
local atan2 = math.atan2

local lighting = {}

lighting.new_light = function(level, x, y, num_rays, distance, angle, width, density, alpha, color, active, special)

	local light = {
		x = x,
		y = y,
		distance = distance,
		angle = angle,
		num_rays = num_rays or 120,
		width = width or pi/4,
		density = density or 5,
		alpha = alpha or 100,
		active = active or true
	}
	
	if color then
		light.r = color.r
		light.g = color.g
		light.b = color.b
	else
		light.r = 255
		light.g = 255
		light.b = 255
	end
	if special == 'spinning' then
		light.type = special

		light.update = function(self, dt)
			self.angle = self.angle + dt
		end
	else
		light.type = 'static'
		light.update = function(self) do return end end
	end 
	
	table.insert(level.lights, light)
	
end --lighting.new

lighting.update = function(dt, level)

	for _, light in pairs(level.lights) do
		if light.active then
			light:update(dt)
		end 
	end --for light
end --lighting.update

lighting.update_canvas = function(canvas, level)
	
	canvas:clear(50, 50, 50, 200)
	_lg.setCanvas(canvas)
	_lg.setBlendMode("alpha")
	
	for _, light in pairs(level.lights) do
		local dist,x_hit,y_hit
		_lg.setColor(light.r,light.g,light.b,light.alpha)
		_lg.setLine(light.density, "rough")
		
		for ray = -1*light.num_rays, light.num_rays do
			this_angle = (ray/light.num_rays)*light.width + light.angle
			x_hit, y_hit, dist = castSingleRay(light.x/tileSize, light.y/tileSize, level, this_angle) 
			if dist then
				if dist*tileSize > light.distance then
					x_hit = light.distance*cos(this_angle) + light.x
					y_hit = light.distance*sin(this_angle) + light.y
				end
				_lg.line(light.x-mapX, light.y-mapY, x_hit-mapX, y_hit-mapY)	
			end
			--print(light.x,light.y,dist, "at end")
		end --ray	
		
		_lg.setColor(255,255,255,255)
	end --light

	_lg.setCanvas()
	
end --lighting.draw

function castSingleRay(lightx, lighty, level, rayAngle1) 

	local twoPI = 2*pi
	-- first make sure the angle is between 0 and 360 degrees
	local rayAngle = rayAngle1 % twoPI;
	if (rayAngle < 0) then rayAngle = rayAngle + twoPI end

	-- moving right/left? up/down? Determined by which quadrant the angle is in.
	local right = (rayAngle > twoPI * 0.75) or (rayAngle < twoPI * 0.25);
	local up = (rayAngle < 0) or (rayAngle > pi);

	-- only do these once
	local angleSin = sin(rayAngle);
	local angleCos = cos(rayAngle);

	local vHit = false
	local dist = 0;	-- the distance to the block we hit
	local xHit = 0; 	-- the x and y coord of where the ray hit the block
	local yHit = 0;

	local x1
	local y1
	
	local wallX;	-- the (x,y) map coords of the block
	local wallY

	local slope = angleSin / angleCos;
	local dX
    if right then dX = 1 else dX=-1 end
    
    local dY = dX * slope; 		-- how much to move up or down

	local x 
    if right then
        x = math.ceil(lightx) 
    else
        x = math.floor(lightx);
    end
    
	local y = lighty + (x - lightx) * slope;
	while (x >= 0 and x <= level.w  and y >= 0 and y <= level.h ) do

		local wallX 
        if right then
            wallX= math.floor(x);
        else
            wallX = math.floor(x-1);
        end
		local wallY = math.floor(y);
		-- is this point inside a wall block?
		if level.grid[wallY][wallX].solid then 
			local distX = x - lightx;
			local distY = y - lighty;
			dist = distX*distX + distY*distY;

			xHit = x;
			yHit = y;
			vHit = true
			x1 = wallX
			y1 = wallY

			break
		end

		x = x+dX;
		y = y+dY;
		
	end

	local slope = angleCos / angleSin;
	local dY = up and -1 or 1;

	local dX = dY * slope;

	local y 
    
    if up then
        y= math.floor(lighty) 
    else
        y= math.ceil(lighty);
    end

	local x = lightx + (y - lighty) * slope;
	while (x >= 0 and x <= level.w  and y >= 0 and y <= level.h) do

		local wallY
        if up then
            wallY= math.floor(y-1);
        else
            wallY= math.floor(y);
        end
		local wallX = math.floor(x);
	
		if level.grid[wallY][wallX].solid then
			local distX = x - lightx;
			local distY = y - lighty;
			local blockDist = distX*distX + distY*distY;

			if (dist==0 or blockDist < dist) then

				dist = blockDist;
				xHit = x;
				yHit = y;
				vHit = false;
	
				x1 = wallX
				y1 = wallY
			end

			break;

		end

		x = x+dX;
		y = y+dY;
	end
	
	if (dist>0) then
		dist =  math.sqrt(dist)
		local extra = 0.35
		--these extra ones goes a little bit further, so that walls are visible
		xHit_extra = angleCos*(dist+extra) + lightx
		yHit_extra = angleSin*(dist+extra) + lighty
		return xHit_extra*tileSize, yHit_extra*tileSize, dist
	end

end

return lighting
